@@ -12,6 +12,7 @@ type ResponseFormatter interface {
 	Format(writer io.Writer, data []byte) error
 	Title() string
 	Searchable() bool
+	Search(string, []byte) ([]string, error)
 }
 
 func New(appConfig *config.Config, contentType string) ResponseFormatter {
@@ -23,6 +24,6 @@ func New(appConfig *config.Config, contentType string) ResponseFormatter {
 	} else if strings.Index(contentType, "text") == -1 && strings.Index(contentType, "application") == -1 {
 		return &binaryFormatter{}
 	} else {
-		return &textFormatter{}
+		return &TextFormatter{}
 	}
 }
