@@ -4,14 +4,17 @@ import (
 	"bytes"
 	"compress/gzip"
 	"crypto/tls"
+	"encoding/json"
 	"errors"
 	"fmt"
 	"io"
 	"io/ioutil"
 	"log"
+	"mime/multipart"
 	"net/http"
 	"net/url"
 	"os"
+	"path"
 	"regexp"
 	"runtime"
 	"sort"
@@ -26,9 +29,10 @@ import (
 
 	"github.com/jroimartin/gocui"
 	"github.com/mattn/go-runewidth"
+	"github.com/nsf/termbox-go"
 )
 
-const VERSION = "0.3.0"
+const VERSION = "0.4.0"
 
 const TIMEOUT_DURATION = 5 // in seconds
 const WINDOWS_OS = "windows"
@@ -47,24 +51,30 @@ const (
 	RESPONSE_HEADERS_VIEW = "response-headers"
 	RESPONSE_BODY_VIEW    = "response-body"
 
-	SEARCH_PROMPT_VIEW = "prompt"
-	POPUP_VIEW         = "popup_view"
-	AUTOCOMPLETE_VIEW  = "autocomplete_view"
-	ERROR_VIEW         = "error_view"
-	HISTORY_VIEW       = "history"
-	SAVE_DIALOG_VIEW   = "save-dialog"
-	SAVE_RESULT_VIEW   = "save-result"
-	METHOD_LIST_VIEW   = "method-list"
-	HELP_VIEW          = "help"
+	SEARCH_PROMPT_VIEW        = "prompt"
+	POPUP_VIEW                = "popup_view"
+	AUTOCOMPLETE_VIEW         = "autocomplete_view"
+	ERROR_VIEW                = "error_view"
+	HISTORY_VIEW              = "history"
+	SAVE_DIALOG_VIEW          = "save-dialog"
+	SAVE_RESPONSE_DIALOG_VIEW = "save-response-dialog"
+	LOAD_REQUEST_DIALOG_VIEW  = "load-request-dialog"
+	SAVE_REQUEST_DIALOG_VIEW  = "save-request-dialog"
+	SAVE_RESULT_VIEW          = "save-result"
+	METHOD_LIST_VIEW          = "method-list"
+	HELP_VIEW                 = "help"
 )
 
 var VIEW_TITLES = map[string]string{
-	POPUP_VIEW:       "Info",
-	ERROR_VIEW:       "Error",
-	HISTORY_VIEW:     "History",
-	SAVE_DIALOG_VIEW: "Save Response (enter to submit, ctrl+q to cancel)",
-	METHOD_LIST_VIEW: "Methods",
-	HELP_VIEW:        "Help",
+	POPUP_VIEW:                "Info",
+	ERROR_VIEW:                "Error",
+	HISTORY_VIEW:              "History",
+	SAVE_RESPONSE_DIALOG_VIEW: "Save Response (enter to submit, ctrl+q to cancel)",
+	LOAD_REQUEST_DIALOG_VIEW:  "Load Request (enter to submit, ctrl+q to cancel)",
+	SAVE_REQUEST_DIALOG_VIEW:  "Save Request (enter to submit, ctrl+q to cancel)",
+	SAVE_RESULT_VIEW:          "Save Result (press enter to close)",
+	METHOD_LIST_VIEW:          "Methods",
+	HELP_VIEW:                 "Help",
 }
 
 type position struct {
@@ -260,6 +270,8 @@ var METHODS = []string{
 
 const DEFAULT_METHOD = http.MethodGet
 
+var DEFAULT_FORMATTER = &formatter.TextFormatter{}
+
 var CLIENT = &http.Client{
 	Timeout: time.Duration(TIMEOUT_DURATION * time.Second),
 }
@@ -302,6 +314,7 @@ type Request struct {
 	RawResponseBody []byte
 	ContentType     string
 	Duration        time.Duration
+	Formatter       formatter.ResponseFormatter
 }
 
 type App struct {
@@ -458,7 +471,7 @@ func (e *AutocompleteEditor) Edit(v *gocui.View, key gocui.Key, ch rune, mod goc
 
 func (e *SearchEditor) Edit(v *gocui.View, key gocui.Key, ch rune, mod gocui.Modifier) {
 	e.wuzzEditor.Edit(v, key, ch, mod)
-	e.wuzzEditor.g.Execute(func(g *gocui.Gui) error {
+	e.wuzzEditor.g.Update(func(g *gocui.Gui) error {
 		e.wuzzEditor.app.PrintBody(g)
 		return nil
 	})
@@ -700,7 +713,7 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 		r.Url = getViewValue(g, URL_VIEW)
 		u, err := url.Parse(r.Url)
 		if err != nil {
-			g.Execute(func(g *gocui.Gui) error {
+			g.Update(func(g *gocui.Gui) error {
 				vrb, _ := g.View(RESPONSE_BODY_VIEW)
 				fmt.Fprintf(vrb, "URL parse error: %v", err)
 				return nil
@@ -710,7 +723,7 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 
 		q, err := url.ParseQuery(strings.Replace(getViewValue(g, URL_PARAMS_VIEW), "\n", "&", -1))
 		if err != nil {
-			g.Execute(func(g *gocui.Gui) error {
+			g.Update(func(g *gocui.Gui) error {
 				vrb, _ := g.View(RESPONSE_BODY_VIEW)
 				fmt.Fprintf(vrb, "Invalid GET parameters: %v", err)
 				return nil
@@ -735,7 +748,7 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 			if header != "" {
 				header_parts := strings.SplitN(header, ": ", 2)
 				if len(header_parts) != 2 {
-					g.Execute(func(g *gocui.Gui) error {
+					g.Update(func(g *gocui.Gui) error {
 						vrb, _ := g.View(RESPONSE_BODY_VIEW)
 						fmt.Fprintf(vrb, "Invalid header: %v", header)
 						return nil
@@ -751,16 +764,58 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 		// parse POST/PUT/PATCH data
 		if r.Method == http.MethodPost || r.Method == http.MethodPut || r.Method == http.MethodPatch {
 			bodyStr := getViewValue(g, REQUEST_DATA_VIEW)
-			if headers.Get("Content-Type") == "application/x-www-form-urlencoded" {
-				bodyStr = strings.Replace(bodyStr, "\n", "&", -1)
+			if headers.Get("Content-Type") != "multipart/form-data" {
+				if headers.Get("Content-Type") == "application/x-www-form-urlencoded" {
+					bodyStr = strings.Replace(bodyStr, "\n", "&", -1)
+				}
+				body = bytes.NewBufferString(bodyStr)
+			} else {
+				var bodyBytes bytes.Buffer
+				multiWriter := multipart.NewWriter(&bodyBytes)
+				defer multiWriter.Close()
+				postData, err := url.ParseQuery(strings.Replace(getViewValue(g, REQUEST_DATA_VIEW), "\n", "&", -1))
+				if err != nil {
+					return err
+				}
+				for postKey, postValues := range postData {
+					for i := range postValues {
+						if len([]rune(postValues[i])) > 0 && postValues[i][0] == '@' {
+							file, err := os.Open(postValues[i][1:])
+							if err != nil {
+								g.Update(func(g *gocui.Gui) error {
+									vrb, _ := g.View(RESPONSE_BODY_VIEW)
+									fmt.Fprintf(vrb, "Error: %v", err)
+									return nil
+								})
+								return err
+							}
+							defer file.Close()
+							fw, err := multiWriter.CreateFormFile(postKey, path.Base(postValues[i][1:]))
+							if err != nil {
+								return err
+							}
+							if _, err := io.Copy(fw, file); err != nil {
+								return err
+							}
+						} else {
+							fw, err := multiWriter.CreateFormField(postKey)
+							if err != nil {
+								return err
+							}
+							if _, err := fw.Write([]byte(postValues[i])); err != nil {
+								return err
+							}
+						}
+					}
+				}
+				body = bytes.NewReader(bodyBytes.Bytes())
 			}
-			body = bytes.NewBufferString(bodyStr)
 		}
 
 		// create request
 		req, err := http.NewRequest(r.Method, u.String(), body)
 		if err != nil {
-			g.Execute(func(g *gocui.Gui) error {
+			g.Update(func(g *gocui.Gui) error {
 				vrb, _ := g.View(RESPONSE_BODY_VIEW)
 				fmt.Fprintf(vrb, "Request error: %v", err)
 				return nil
@@ -769,12 +824,17 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 		}
 		req.Header = headers
 
+		// set the `Host` header
+		if headers.Get("Host") != "" {
+			req.Host = headers.Get("Host")
+		}
+
 		// do request
 		start := time.Now()
 		response, err := CLIENT.Do(req)
 		r.Duration = time.Since(start)
 		if err != nil {
-			g.Execute(func(g *gocui.Gui) error {
+			g.Update(func(g *gocui.Gui) error {
 				vrb, _ := g.View(RESPONSE_BODY_VIEW)
 				fmt.Fprintf(vrb, "Response error: %v", err)
 				return nil
@@ -791,7 +851,7 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 				defer reader.Close()
 				response.Body = reader
 			} else {
-				g.Execute(func(g *gocui.Gui) error {
+				g.Update(func(g *gocui.Gui) error {
 					vrb, _ := g.View(RESPONSE_BODY_VIEW)
 					fmt.Fprintf(vrb, "Cannot uncompress response: %v", err)
 					return nil
@@ -805,12 +865,14 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 			r.RawResponseBody = bodyBytes
 		}
 
+		r.Formatter = formatter.New(a.config, r.ContentType)
+
 		// add to history
 		a.history = append(a.history, r)
 		a.historyIndex = len(a.history) - 1
 
 		// render response
-		g.Execute(func(g *gocui.Gui) error {
+		g.Update(func(g *gocui.Gui) error {
 			vrh, _ := g.View(RESPONSE_HEADERS_VIEW)
 
 			a.PrintBody(g)
@@ -848,7 +910,7 @@ func (a *App) SubmitRequest(g *gocui.Gui, _ *gocui.View) error {
 }
 
 func (a *App) PrintBody(g *gocui.Gui) {
-	g.Execute(func(g *gocui.Gui) error {
+	g.Update(func(g *gocui.Gui) error {
 		if len(a.history) == 0 {
 			return nil
 		}
@@ -859,27 +921,32 @@ func (a *App) PrintBody(g *gocui.Gui) {
 		vrb, _ := g.View(RESPONSE_BODY_VIEW)
 		vrb.Clear()
 
-		responseFormatter := formatter.New(a.config, req.ContentType)
+		var responseFormatter formatter.ResponseFormatter
+		responseFormatter = req.Formatter
+
 		vrb.Title = VIEW_PROPERTIES[vrb.Name()].title + " " + responseFormatter.Title()
 
 		search_text := getViewValue(g, "search")
 		if search_text == "" || !responseFormatter.Searchable() {
 			err := responseFormatter.Format(vrb, req.RawResponseBody)
 			if err != nil {
-				return err
+				fmt.Fprintf(vrb, "Error: cannot decode response body: %v", err)
+				return nil
 			}
 			if _, err := vrb.Line(0); !a.config.General.PreserveScrollPosition || err != nil {
 				vrb.SetOrigin(0, 0)
 			}
 			return nil
 		}
+		if !a.config.General.ContextSpecificSearch {
+			responseFormatter = DEFAULT_FORMATTER
+		}
 		vrb.SetOrigin(0, 0)
-		search_re, err := regexp.Compile(search_text)
+		results, err := responseFormatter.Search(search_text, req.RawResponseBody)
 		if err != nil {
-			fmt.Fprint(vrb, "Error: invalid search regexp")
+			fmt.Fprint(vrb, "Search error: ", err)
 			return nil
 		}
-		results := search_re.FindAll(req.RawResponseBody, 1000)
 		if len(results) == 0 {
 			vrb.Title = "No results"
 			fmt.Fprint(vrb, "Error: no results")
@@ -1055,50 +1122,6 @@ func (a *App) SetKeys(g *gocui.Gui) error {
 		return nil
 	})
 
-	g.SetKeybinding(SAVE_DIALOG_VIEW, gocui.KeyEnter, gocui.ModNone, func(g *gocui.Gui, v *gocui.View) error {
-		defer a.closePopup(g, SAVE_DIALOG_VIEW)
-
-		saveLocation := getViewValue(g, SAVE_DIALOG_VIEW)
-
-		if len(a.history) == 0 {
-			return nil
-		}
-		req := a.history[a.historyIndex]
-		if req.RawResponseBody == nil {
-			return nil
-		}
-
-		err := ioutil.WriteFile(saveLocation, req.RawResponseBody, 0644)
-
-		var saveResult string
-		if err == nil {
-			saveResult = "Response saved successfully."
-		} else {
-			saveResult = "Error saving response: " + err.Error()
-		}
-
-		popupTitle := "Save Result (press enter to close)"
-
-		saveResHeight := 1
-		saveResWidth := len(saveResult) + 1
-		if len(popupTitle)+2 > saveResWidth {
-			saveResWidth = len(popupTitle) + 2
-		}
-		maxX, _ := g.Size()
-		if saveResWidth > maxX {
-			saveResHeight = saveResWidth/maxX + 1
-			saveResWidth = maxX
-		}
-
-		saveResultPopup, err := a.CreatePopupView(SAVE_RESULT_VIEW, saveResWidth, saveResHeight, g)
-		saveResultPopup.Title = popupTitle
-		setViewTextAndCursor(saveResultPopup, saveResult)
-		g.SetViewOnTop(SAVE_RESULT_VIEW)
-		g.SetCurrentView(SAVE_RESULT_VIEW)
-
-		return err
-	})
-
 	g.SetKeybinding(SAVE_DIALOG_VIEW, gocui.KeyCtrlQ, gocui.ModNone, func(g *gocui.Gui, v *gocui.View) error {
 		a.closePopup(g, SAVE_DIALOG_VIEW)
 		return nil
@@ -1148,6 +1171,63 @@ func (a *App) CreatePopupView(name string, width, height int, g *gocui.Gui) (v *
 	return
 }
 
+func (a *App) LoadRequest(g *gocui.Gui, loadLocation string) (err error) {
+	requestJson, ioErr := ioutil.ReadFile(loadLocation)
+	if ioErr != nil {
+		g.Update(func(g *gocui.Gui) error {
+			vrb, _ := g.View(RESPONSE_BODY_VIEW)
+			vrb.Clear()
+			fmt.Fprintf(vrb, "File reading error: %v", ioErr)
+			return nil
+		})
+		return nil
+	}
+
+	var requestMap map[string]string
+	jsonErr := json.Unmarshal(requestJson, &requestMap)
+	if jsonErr != nil {
+		g.Update(func(g *gocui.Gui) error {
+			vrb, _ := g.View(RESPONSE_BODY_VIEW)
+			vrb.Clear()
+			fmt.Fprintf(vrb, "JSON decoding error: %v", jsonErr)
+			return nil
+		})
+		return nil
+	}
+
+	var v *gocui.View
+	url, exists := requestMap[URL_VIEW]
+	if exists {
+		v, _ = g.View(URL_VIEW)
+		setViewTextAndCursor(v, url)
+	}
+
+	method, exists := requestMap[REQUEST_METHOD_VIEW]
+	if exists {
+		v, _ = g.View(REQUEST_METHOD_VIEW)
+		setViewTextAndCursor(v, method)
+	}
+
+	params, exists := requestMap[URL_PARAMS_VIEW]
+	if exists {
+		v, _ = g.View(URL_PARAMS_VIEW)
+		setViewTextAndCursor(v, params)
+	}
+
+	data, exists := requestMap[REQUEST_DATA_VIEW]
+	if exists {
+		v, _ = g.View(REQUEST_DATA_VIEW)
+		setViewTextAndCursor(v, data)
+	}
+
+	headers, exists := requestMap[REQUEST_HEADERS_VIEW]
+	if exists {
+		v, _ = g.View(REQUEST_HEADERS_VIEW)
+		setViewTextAndCursor(v, headers)
+	}
+	return nil
+}
+
 func (a *App) ToggleHistory(g *gocui.Gui, _ *gocui.View) (err error) {
 	// Destroy if present
 	if a.currentPopup == HISTORY_VIEW {
@@ -1211,15 +1291,14 @@ func (a *App) ToggleMethodList(g *gocui.Gui, _ *gocui.View) (err error) {
 	return
 }
 
-func (a *App) OpenSaveDialog(g *gocui.Gui, _ *gocui.View) (err error) {
+func (a *App) OpenSaveDialog(title string, g *gocui.Gui, save func(g *gocui.Gui, v *gocui.View) error) error {
 	dialog, err := a.CreatePopupView(SAVE_DIALOG_VIEW, 60, 1, g)
 	if err != nil {
-		return
+		return err
 	}
-
 	g.Cursor = true
 
-	dialog.Title = VIEW_TITLES[SAVE_DIALOG_VIEW]
+	dialog.Title = title
 	dialog.Editable = true
 	dialog.Wrap = false
 
@@ -1234,7 +1313,30 @@ func (a *App) OpenSaveDialog(g *gocui.Gui, _ *gocui.View) (err error) {
 	g.SetViewOnTop(SAVE_DIALOG_VIEW)
 	g.SetCurrentView(SAVE_DIALOG_VIEW)
 	dialog.SetCursor(0, len(currentDir))
-	return
+	g.DeleteKeybinding(SAVE_DIALOG_VIEW, gocui.KeyEnter, gocui.ModNone)
+	g.SetKeybinding(SAVE_DIALOG_VIEW, gocui.KeyEnter, gocui.ModNone, save)
+	return nil
+}
+
+func (a *App) OpenSaveResultView(saveResult string, g *gocui.Gui) (err error) {
+	popupTitle := VIEW_TITLES[SAVE_RESULT_VIEW]
+	saveResHeight := 1
+	saveResWidth := len(saveResult) + 1
+	if len(popupTitle)+2 > saveResWidth {
+		saveResWidth = len(popupTitle) + 2
+	}
+	maxX, _ := g.Size()
+	if saveResWidth > maxX {
+		saveResHeight = saveResWidth/maxX + 1
+		saveResWidth = maxX
+	}
+
+	saveResultPopup, err := a.CreatePopupView(SAVE_RESULT_VIEW, saveResWidth, saveResHeight, g)
+	saveResultPopup.Title = popupTitle
+	setViewTextAndCursor(saveResultPopup, saveResult)
+	g.SetViewOnTop(SAVE_RESULT_VIEW)
+	g.SetCurrentView(SAVE_RESULT_VIEW)
+	return err
 }
 
 func (a *App) restoreRequest(g *gocui.Gui, idx int) {
@@ -1316,6 +1418,7 @@ func (a *App) ParseArgs(g *gocui.Gui, args []string) error {
 	arg_index := 1
 	args_len := len(args)
 	accept_types := make([]string, 0, 8)
+	var body_data []string
 	for arg_index < args_len {
 		arg := args[arg_index]
 		switch arg {
@@ -1326,7 +1429,7 @@ func (a *App) ParseArgs(g *gocui.Gui, args []string) error {
 			arg_index += 1
 			header := args[arg_index]
 			fmt.Fprintf(vheader, "%v\n", header)
-		case "-d", "--data", "--data-binary":
+		case "-d", "--data", "--data-binary", "--data-urlencode":
 			if arg_index == args_len-1 {
 				return errors.New("No POST/PUT/PATCH value specified")
 			}
@@ -1334,14 +1437,19 @@ func (a *App) ParseArgs(g *gocui.Gui, args []string) error {
 			arg_index += 1
 			set_data = true
 			set_binary_data = arg == "--data-binary"
+			arg_data := args[arg_index]
 
-			data := args[arg_index]
 			if !set_binary_data {
-				data, _ = url.QueryUnescape(data)
 				content_type = "form"
 			}
-			vdata, _ := g.View(REQUEST_DATA_VIEW)
-			setViewTextAndCursor(vdata, data)
+
+			if arg == "--data-urlencode" {
+				// TODO: Replace with `url.PathEscape(..)` in Go 1.8
+				arg_data_url := &url.URL{Path: arg_data}
+				arg_data = arg_data_url.String()
+			}
+
+			body_data = append(body_data, arg_data)
 		case "-j", "--json":
 			if arg_index == args_len-1 {
 				return errors.New("No POST/PUT/PATCH value specified")
@@ -1381,6 +1489,12 @@ func (a *App) ParseArgs(g *gocui.Gui, args []string) error {
 			if !strings.Contains(getViewValue(g, REQUEST_HEADERS_VIEW), "Accept-Encoding") {
 				fmt.Fprintln(vh, "Accept-Encoding: gzip, deflate")
 			}
+		case "-e", "--editor":
+			if arg_index == args_len-1 {
+				return errors.New("No timeout value specified")
+			}
+			arg_index += 1
+			a.config.General.Editor = args[arg_index]
 		case "-k", "--insecure":
 			a.config.General.Insecure = true
 		case "-R", "--disable-redirects":
@@ -1440,10 +1554,28 @@ func (a *App) ParseArgs(g *gocui.Gui, args []string) error {
 			default:
 				return errors.New("Unknown proxy protocol")
 			}
+		case "-F", "--form":
+			if arg_index == args_len-1 {
+				return errors.New("No POST/PUT/PATCH value specified")
+			}
+
+			arg_index += 1
+			form_str := args[arg_index]
+			content_type = "multipart"
+			set_data = true
+			vdata, _ := g.View(REQUEST_DATA_VIEW)
+			setViewTextAndCursor(vdata, form_str)
+		case "-f", "--file":
+			if arg_index == args_len-1 {
+				return errors.New("-f or --file requires a file path be provided as an argument")
+			}
+			arg_index += 1
+			loadLocation := args[arg_index]
+			a.LoadRequest(g, loadLocation)
 		default:
 			u := args[arg_index]
 			if strings.Index(u, "http://") != 0 && strings.Index(u, "https://") != 0 {
-				u = "http://" + u
+				u = fmt.Sprintf("%v://%v", a.config.General.DefaultURLScheme, u)
 			}
 			parsed_url, err := url.Parse(u)
 			if err != nil || parsed_url.Host == "" {
@@ -1476,6 +1608,14 @@ func (a *App) ParseArgs(g *gocui.Gui, args []string) error {
 		fmt.Fprintf(vheader, "Accept: %v\n", strings.Join(accept_types, ","))
 	}
 
+	var merged_body_data string
+	if set_data && !set_binary_data {
+		merged_body_data = strings.Join(body_data, "&")
+	}
+
+	vdata, _ := g.View(REQUEST_DATA_VIEW)
+	setViewTextAndCursor(vdata, merged_body_data)
+
 	return nil
 }
 
@@ -1545,6 +1685,10 @@ Usage: wuzz [-H|--header HEADER]... [-d|--data|--data-binary DATA] [-X|--request
 
 Other command line options:
   -c, --config PATH        Specify custom configuration file
+  -e, --editor EDITOR      Specify external editor command
+  -f, --file REQUEST       Load a previous request
+  -F, --form DATA          Add multipart form request data and set related request headers
+                           If the value starts with @ it will be handled as a file path for upload
   -h, --help               Show this
   -j, --json JSON          Add JSON request data and set related request headers
   -k, --insecure           Allow insecure SSL certs
@@ -1562,6 +1706,8 @@ Other command line options:
 Key bindings:
   ctrl+r              Send request
   ctrl+s              Save response
+  ctrl+e              Save request
+  ctrl+f              Load request
   tab, ctrl+j         Next window
   shift+tab, ctrl+k   Previous window
   alt+h               Show history
@@ -1589,10 +1735,18 @@ func main() {
 			}
 		}
 	}
-	g, err := gocui.NewGui(gocui.Output256)
+	var g *gocui.Gui
+	var err error
+	for _, outputMode := range []gocui.OutputMode{gocui.Output256, gocui.OutputNormal, gocui.OutputMode(termbox.OutputGrayscale)} {
+		g, err = gocui.NewGui(outputMode)
+		if err == nil {
+			break
+		}
+	}
 	if err != nil {
 		log.Panicln(err)
 	}
+
 	if runtime.GOOS == WINDOWS_OS && runewidth.IsEastAsian() {
 		g.ASCII = true
 	}
